from typing import Any, Dict, List


def extract_anthropic_tool_calls(
    response_data: Dict[str, Any],
) -> List[Dict[str, Any]]:
    """
    Extract tool calls from Anthropic API response.

    Args:
        response_data: The raw response data from Anthropic API

    Returns:
        List of tool call dictionaries with standardized format
    """
    tool_calls = []

    # Handle None response (e.g., when request is cancelled)
    if response_data is None:
        return tool_calls

    # Check if response has content array
    content = response_data.get("content", [])
    if not isinstance(content, list):
        return tool_calls

    for content_block in content:
        if (
            isinstance(content_block, dict)
            and content_block.get("type") == "tool_use"
        ):
            tool_call = {
                "id": content_block.get("id", ""),
                "name": content_block.get("name", ""),
                "input": content_block.get("input", {}),
                "type": "tool_use",
            }
            tool_calls.append(tool_call)

    return tool_calls


def extract_anthropic_tool_results(
    response_data: Dict[str, Any],
) -> List[Dict[str, Any]]:
    """
    Extract tool results from Anthropic API response (for tool result blocks).

    Args:
        response_data: The raw response data from Anthropic API

    Returns:
        List of tool result dictionaries
    """
    tool_results = []

    # Handle None response (e.g., when request is cancelled)
    if response_data is None:
        return tool_results

    # Check if response has content array
    content = response_data.get("content", [])
    if not isinstance(content, list):
        return tool_results

    for content_block in content:
        if (
            isinstance(content_block, dict)
            and content_block.get("type") == "tool_result"
        ):
            tool_result = {
                "tool_use_id": content_block.get("tool_use_id", ""),
                "content": content_block.get("content", ""),
                "is_error": content_block.get("is_error", False),
                "type": "tool_result",
            }
            tool_results.append(tool_result)

    return tool_results


def extract_anthropic_text_content(response_data: Dict[str, Any]) -> str:
    """
    Extract text content from Anthropic API response, filtering out tool calls.

    Args:
        response_data: The raw response data from Anthropic API

    Returns:
        Concatenated text content
    """
    text_content = ""

    # Handle None response (e.g., when request is cancelled)
    if response_data is None:
        return text_content

    # Check if response has content array
    content = response_data.get("content", [])
    if not isinstance(content, list):
        return text_content

    for content_block in content:
        if (
            isinstance(content_block, dict)
            and content_block.get("type") == "text"
        ):
            text_content += content_block.get("text", "")

    return text_content


def get_anthropic_response_summary(
    response_data: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Get a comprehensive summary of Anthropic API response including text and tool calls.

    Args:
        response_data: The raw response data from Anthropic API

    Returns:
        Dictionary containing text content, tool calls, and summary info
    """
    # Handle None response (e.g., when request is cancelled)
    if response_data is None:
        return {
            "text_content": "",
            "tool_calls": [],
            "tool_results": [],
            "has_tool_calls": False,
            "has_tool_results": False,
        }

    tool_calls = extract_anthropic_tool_calls(response_data)
    tool_results = extract_anthropic_tool_results(response_data)
    text_content = extract_anthropic_text_content(response_data)

    return {
        "text_content": text_content,
        "tool_calls": tool_calls,
        "tool_results": tool_results,
        "has_tool_calls": len(tool_calls) > 0,
        "has_tool_results": len(tool_results) > 0,
        "tool_call_count": len(tool_calls),
        "tool_result_count": len(tool_results),
    }


def extract_anthropic_response_with_tools(response_data: Dict[str, Any]):
    """
    Enhanced response extractor for Anthropic that handles both text and tool calls.

    Args:
        response_data: The raw response data from Anthropic API

    Returns:
        Dictionary containing text content and tool call information, or just text for backward compatibility
    """
    response_summary = get_anthropic_response_summary(response_data)

    # Combine text content and tool calls for comprehensive output
    output_content = {
        "text": response_summary["text_content"],
        "tool_calls": response_summary["tool_calls"],
        "tool_results": response_summary["tool_results"],
    }

    # If no tool calls, just return the text for backward compatibility
    if (
        not response_summary["has_tool_calls"]
        and not response_summary["has_tool_results"]
    ):
        return response_summary["text_content"]

    return output_content



from contextvars import ContextVar
from typing import Optional

from fastapi import Request
from langfuse.client import StatefulTraceClient

request_context: ContextVar[Request] = ContextVar(
    "request_context", default=None
)
tracer_context: ContextVar[StatefulTraceClient] = ContextVar(
    "tracer_context", default=None
)
request_metadata_context: ContextVar[dict] = ContextVar(
    "request_metadata_context", default={}
)
generation_id_context: ContextVar[str] = ContextVar(
    "generation_id_context", default=None
)


def get_current_generation_id() -> Optional[str]:
    """
    Get the current generation ID from context.

    Returns:
        str: The current generation ID if available, None otherwise
    """
    return generation_id_context.get()


def set_generation_id(generation_id: str) -> None:
    """
    Set the generation ID in context.

    Args:
        generation_id: The generation ID to store in context
    """
    generation_id_context.set(generation_id)


def clear_generation_id() -> None:
    """
    Clear the generation ID from context.
    """
    generation_id_context.set(None)



import json
from datetime import datetime
from unittest.mock import MagicMock, mock_open, patch

import pytest

from observe_traces.config.context_util import (
    request_context,
    request_metadata_context,
    tracer_context,
)
from observe_traces.config.langfuse_init import LangfuseInitializer
from observe_traces.config.langfuse_service import _LangfuseService
from observe_traces.tracer.embed_tracer import (
    calculate_pinecone_price,
    embedding_tracing,
)
from observe_traces.tracer.llm_tracer import (
    llm_tracing,
    parse_groq_tokens,
    parse_openai_tokens,
)
from observe_traces.tracer.rerank_tracer import (
    calculate_pinecone_rerank_price,
    reranking_tracing,
)
from observe_traces.tracer.vector_tracer import (
    calculate_pinecone_price as calculate_pinecone_vectordb_price,
)
from observe_traces.tracer.vector_tracer import vectordb_tracing
from observe_traces.utils.token_costs import get_token_costs

# Test data
TEST_PUBLIC_KEY = "pk-lf-ca85952c-d5cf-4d21-8fff-fa2942a0e33b"
TEST_SECRET_KEY = "sk-lf-d1b12ef9-077e-431d-b6ab-1b764f744217"
TEST_HOST = "http://localhost:3000"
TEST_RELEASE = "3.46.0"
TEST_ENVIRONMENT = "unit-testing"

# Mock pricing data
MOCK_PRICING_DATA = {
    "gpt-3.5-turbo": {
        "input_cost_per_token": 0.0000015,
        "output_cost_per_token": 0.000002,
    },
    "claude-3-7-sonnet-20250219": {
        "input_cost_per_token": 0.000003,
        "output_cost_per_token": 0.000015,
    },
    "llama-3.3-70b-versatile": {
        "input_cost_per_token": 0.0000007,
        "output_cost_per_token": 0.0000007,
    },
    "llama-text-embed-v2": {
        "input_cost_per_token": 0.00000016,
        "output_cost_per_token": 0.0,
    },
    "pinecone-rerank-v0": {
        "input_cost_per_token": 0.0001,
        "output_cost_per_token": 0.0,
    },
}


@pytest.fixture
def mock_langfuse_client():
    with patch("observe_traces.config.langfuse_service.Langfuse") as mock:
        client = MagicMock()
        mock.return_value = client
        yield client


@pytest.fixture
def setup_langfuse(mock_langfuse_client):
    LangfuseInitializer.initialize(
        TEST_PUBLIC_KEY,
        TEST_SECRET_KEY,
        TEST_HOST,
        TEST_RELEASE,
        TEST_ENVIRONMENT,
    )
    yield
    LangfuseInitializer.close()


@pytest.fixture
def mock_trace():
    trace = MagicMock()
    tracer_context.set(trace)
    yield trace
    tracer_context.set(None)


@pytest.fixture
def mock_request():
    request = MagicMock()
    request_context.set(request)
    yield request
    request_context.set(None)


@pytest.fixture
def mock_metadata():
    metadata = {"user": "test_user", "apiEndpoint": "test_endpoint"}
    request_metadata_context.set(metadata)
    yield metadata
    request_metadata_context.set({})


# LangfuseInitializer Tests
def test_langfuse_initializer_initialization(mock_langfuse_client):
    LangfuseInitializer.initialize(
        TEST_PUBLIC_KEY,
        TEST_SECRET_KEY,
        TEST_HOST,
        TEST_RELEASE,
        TEST_ENVIRONMENT,
    )
    instance = LangfuseInitializer.get_instance()
    assert instance is not None
    assert instance.langfuse_public_key == TEST_PUBLIC_KEY
    assert instance.langfuse_secret_key == TEST_SECRET_KEY
    assert instance.langfuse_host == TEST_HOST
    assert instance.release == TEST_RELEASE
    assert instance.environment == TEST_ENVIRONMENT
    LangfuseInitializer.close()


def test_langfuse_initializer_singleton(mock_langfuse_client):
    LangfuseInitializer.initialize(
        TEST_PUBLIC_KEY,
        TEST_SECRET_KEY,
        TEST_HOST,
        TEST_RELEASE,
        TEST_ENVIRONMENT,
    )
    instance1 = LangfuseInitializer.get_instance()
    instance2 = LangfuseInitializer.get_instance()
    assert instance1 is instance2
    LangfuseInitializer.close()


def test_langfuse_initializer_close(mock_langfuse_client):
    LangfuseInitializer.initialize(
        TEST_PUBLIC_KEY,
        TEST_SECRET_KEY,
        TEST_HOST,
        TEST_RELEASE,
        TEST_ENVIRONMENT,
    )
    instance = LangfuseInitializer.get_instance()
    assert instance is not None
    LangfuseInitializer.close()
    instance = LangfuseInitializer.get_instance()
    assert instance is None


# Token Costs Tests
def test_get_token_costs():
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(MOCK_PRICING_DATA))
    ):
        # Test OpenAI model
        costs = get_token_costs("gpt-3.5-turbo", "openai")
        assert costs["input_cost_per_token"] == 0.0000015
        assert costs["output_cost_per_token"] == 0.000002

        # Test Anthropic model
        costs = get_token_costs("claude-3-7-sonnet-20250219", "anthropic")
        assert costs["input_cost_per_token"] == 0.000003
        assert costs["output_cost_per_token"] == 0.000015

        # Test Groq model
        costs = get_token_costs("llama-3.3-70b-versatile", "groq")
        assert costs["input_cost_per_token"] == 0.0000007
        assert costs["output_cost_per_token"] == 0.0000007

        # Test embedding model
        costs = get_token_costs("llama-text-embed-v2", "pinecone")
        assert costs["input_cost_per_token"] == 0.00000016
        assert costs["output_cost_per_token"] == 0.0

        # Test reranking model
        costs = get_token_costs("pinecone-rerank-v0", "pinecone")
        assert costs["input_cost_per_token"] == 0.0001
        assert costs["output_cost_per_token"] == 0.0


def test_get_token_costs_invalid_model():
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(MOCK_PRICING_DATA))
    ):
        with pytest.raises(ValueError):
            get_token_costs("invalid-model", "openai")


def test_get_token_costs_invalid_provider():
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(MOCK_PRICING_DATA))
    ):
        with pytest.raises(ValueError):
            get_token_costs("gpt-3.5-turbo", "invalid-provider")


# Price Calculation Tests
def test_calculate_openai_price():
    # Test calculating price for OpenAI model using token costs
    token_costs = get_token_costs("gpt-3.5-turbo", "openai")
    prompt_tokens = 1000
    completion_tokens = 500

    input_price = (prompt_tokens / 1000000) * token_costs[
        "input_cost_per_token"
    ]
    output_price = (completion_tokens / 1000000) * token_costs[
        "output_cost_per_token"
    ]
    total_price = input_price + output_price

    price = {"input": input_price, "output": output_price, "total": total_price}

    assert "input" in price
    assert "output" in price
    assert "total" in price
    assert price["input"] == 0.0015  # 1000 * 0.0000015
    assert price["output"] == 0.001  # 500 * 0.000002
    assert price["total"] == 0.0025  # 0.0015 + 0.001


def test_calculate_anthropic_price():
    # Test calculating price for Anthropic model using token costs
    token_costs = get_token_costs("claude-3-7-sonnet-20250219", "anthropic")
    prompt_tokens = 1000
    completion_tokens = 500

    input_price = (prompt_tokens / 1000000) * token_costs[
        "input_cost_per_token"
    ]
    output_price = (completion_tokens / 1000000) * token_costs[
        "output_cost_per_token"
    ]
    total_price = input_price + output_price

    price = {"input": input_price, "output": output_price, "total": total_price}

    assert "input" in price
    assert "output" in price
    assert "total" in price
    assert price["input"] == 0.003  # 1000 * 0.000003
    assert price["output"] == 0.0075  # 500 * 0.000015
    assert price["total"] == 0.0105  # 0.003 + 0.0075


def test_calculate_groq_price():
    # Test calculating price for Groq model using token costs
    token_costs = get_token_costs("llama-3.3-70b-versatile", "groq")
    prompt_tokens = 1000
    completion_tokens = 500

    input_price = (prompt_tokens / 1000000) * token_costs[
        "input_cost_per_token"
    ]
    output_price = (completion_tokens / 1000000) * token_costs[
        "output_cost_per_token"
    ]
    total_price = input_price + output_price

    price = {"input": input_price, "output": output_price, "total": total_price}

    assert "input" in price
    assert "output" in price
    assert "total" in price
    assert price["input"] == 0.0007  # 1000 * 0.0000007
    assert price["output"] == 0.00035  # 500 * 0.0000007
    assert price["total"] == 0.00105  # 0.0007 + 0.00035


def test_calculate_pinecone_embedding_price():
    price = calculate_pinecone_price("llama-text-embed-v2", 1000)
    assert "tokens" in price
    assert "price_per_1M" in price
    assert "total" in price
    assert price["tokens"] == 1000
    assert price["price_per_1M"] == 0.16
    assert price["total"] > 0


def test_calculate_pinecone_rerank_price():
    price = calculate_pinecone_rerank_price(
        "pinecone-rerank-v0", {"rerank_units": 1000}
    )
    assert "rerank_units" in price
    assert "price_per_1K" in price
    assert "total" in price
    assert price["rerank_units"] == 1000
    assert price["price_per_1K"] == 0.10
    assert price["total"] > 0


def test_calculate_pinecone_vectordb_price():
    price = calculate_pinecone_vectordb_price("read", 1000)
    assert "units" in price
    assert "price" in price
    assert price["units"] == 1000
    assert price["price"] > 0


# Token Parsing Tests
def test_parse_openai_tokens():
    response_data = {
        "usage": {
            "prompt_tokens": 100,
            "completion_tokens": 50,
            "total_tokens": 150,
        }
    }
    tokens = parse_openai_tokens(response_data)
    assert tokens["prompt_tokens"] == 100
    assert tokens["completion_tokens"] == 50
    assert tokens["total_tokens"] == 150


def test_parse_groq_tokens():
    response_data = {
        "usage": {
            "prompt_tokens": 100,
            "completion_tokens": 50,
            "total_tokens": 150,
        }
    }
    tokens = parse_groq_tokens(response_data)
    assert tokens["prompt_tokens"] == 100
    assert tokens["completion_tokens"] == 50
    assert tokens["total_tokens"] == 150


# LLM Tracing Tests
@pytest.mark.asyncio
async def test_llm_tracing(
    setup_langfuse, mock_trace, mock_request, mock_metadata
):
    @llm_tracing("openai")
    async def mock_llm_call(self, **params):
        return {
            "choices": [{"message": {"content": "test response"}}],
            "usage": {
                "prompt_tokens": 100,
                "completion_tokens": 50,
                "total_tokens": 150,
            },
        }

    class MockLLM:
        @llm_tracing("openai")
        async def call(self, **params):
            return await mock_llm_call(self, **params)

    llm = MockLLM()
    result = await llm.call(
        model="gpt-3.5-turbo",
        chat_messages=["test message"],
        system_prompt="test system prompt",
    )
    assert result["choices"][0]["message"]["content"] == "test response"


# Embedding Tracing Tests
@pytest.mark.asyncio
async def test_embedding_tracing(
    setup_langfuse, mock_trace, mock_request, mock_metadata
):
    @embedding_tracing("pinecone")
    async def mock_embed_call(*args, **kwargs):
        return {
            "data": [{"values": [0.1, 0.2, 0.3]}],
            "usage": {"total_tokens": 100},
        }

    result = await mock_embed_call(
        model_name="llama-text-embed-v2",
        inputs=["test text"],
    )
    assert len(result["data"]) == 1
    assert len(result["data"][0]["values"]) == 3


# Reranking Tracing Tests
@pytest.mark.asyncio
async def test_reranking_tracing(
    setup_langfuse, mock_trace, mock_request, mock_metadata
):
    @reranking_tracing("pinecone")
    async def mock_rerank_call(*args, **kwargs):
        return {
            "data": [{"index": 0, "score": 0.9}],
            "usage": {"rerank_units": 1},
        }

    result = await mock_rerank_call(
        model_name="pinecone-rerank-v0",
        query="test query",
        documents=["test document"],
        top_n=1,
    )
    assert len(result["data"]) == 1
    assert result["data"][0]["index"] == 0
    assert result["data"][0]["score"] == 0.9


# Vector DB Tracing Tests
@pytest.mark.asyncio
async def test_vectordb_tracing(
    setup_langfuse, mock_trace, mock_request, mock_metadata
):
    @vectordb_tracing("pinecone", "read")
    async def mock_vectordb_call(*args, **kwargs):
        return {
            "matches": [{"id": "1", "score": 0.9}],
            "usage": {"read_units": 1},
        }

    result = await mock_vectordb_call(
        query=[0.1, 0.2, 0.3],
        top_k=1,
    )
    assert len(result["matches"]) == 1
    assert result["matches"][0]["id"] == "1"
    assert result["matches"][0]["score"] == 0.9


# LangfuseService Tests
@pytest.mark.asyncio
async def test_create_generation_for_llm(setup_langfuse, mock_trace):
    generation_data = {
        "model_name": "gpt-3.5-turbo",
        "service_provider": "openai",
        "input": {
            "user": ["test message"],
            "system": "test system prompt",
        },
        "output": "test response",
        "usage": {
            "prompt_tokens": 100,
            "completion_tokens": 50,
            "total_tokens": 150,
        },
        "cost_details": {
            "input": 0.1,
            "output": 0.2,
            "total": 0.3,
        },
        "start_time": datetime.now(),
        "end_time": datetime.now(),
    }
    generation_id = await _LangfuseService.create_generation_for_LLM(
        "test_trace_id", generation_data, "test_name"
    )
    assert generation_id is not None


@pytest.mark.asyncio
async def test_create_span_for_embedding(setup_langfuse, mock_trace):
    span_data = {
        "service_provider": "pinecone",
        "model_name": "llama-text-embed-v2",
        "input": ["test text"],
        "tokens": {"total_tokens": 100},
        "price": {"total": 0.1},
        "input_count": 1,
        "response_time": 0.1,
        "start_time": datetime.now(),
        "end_time": datetime.now(),
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "embedding_dimensions": 3,
    }
    span_id = await _LangfuseService.create_span_for_embedding(
        "test_trace_id", span_data, "test_name"
    )
    assert span_id is not None


@pytest.mark.asyncio
async def test_create_span_for_reranking(setup_langfuse, mock_trace):
    span_data = {
        "service_provider": "pinecone",
        "model_name": "pinecone-rerank-v0",
        "tokens": {"rerank_units": 1},
        "price": {"total": 0.1},
        "query": "test query",
        "documents": ["test document"],
        "document_count": 1,
        "top_n": 1,
        "rerank_results": ["test document"],
        "response_time": 0.1,
        "start_time": datetime.now(),
        "end_time": datetime.now(),
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    }
    span_id = await _LangfuseService.create_span_for_reranking(
        "test_trace_id", span_data, "test_name"
    )
    assert span_id is not None


@pytest.mark.asyncio
async def test_create_span_for_vectordb(setup_langfuse, mock_trace):
    span_data = {
        "service_provider": "pinecone",
        "operation_type": "read",
        "response": [{"id": "1", "score": 0.9}],
        "operation_details": {"top_k": 1},
        "units": 1,
        "price": 0.1,
        "query": [0.1, 0.2, 0.3],
        "start_time": datetime.now(),
        "end_time": datetime.now(),
        "response_time": 0.1,
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    }
    span_id = await _LangfuseService.create_span_for_vectorDB(
        "test_trace_id", span_data, "test_name"
    )
    assert span_id is not None



from .context_util import (
    request_context,
    request_metadata_context,
    tracer_context,
)
from .langfuse_init import LangfuseInitializer
from .langfuse_service import _LangfuseService

__all__ = [
    "LangfuseInitializer",
    "_LangfuseService",
    "request_context",
    "tracer_context",
    "request_metadata_context",
    "LangfuseInitializer",
]



"""
OOP vector database tracer implementation.
"""

from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, List, Optional

from ..base.tracer import BaseTracer
from ..config.context_util import request_context
from ..config.langfuse_service import _LangfuseService
from ..utils.tracer_utils import log_error


class VectorDBTracer(BaseTracer):
    """OOP implementation of vector database tracing."""

    def __init__(
        self,
        provider,
        operation_type: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        super().__init__(provider, variable_mapping, metadata_config)
        self.operation_type = operation_type

    async def trace(self, func: Callable, *args, **kwargs) -> Any:
        """Execute vector database tracing for the function call."""
        if not self.should_trace():
            return await func(*args, **kwargs)

        # STEP 1: Always execute the function first
        try:
            result = await func(*args, **kwargs)
        except Exception as function_error:
            log_error(function_error, "Original function failed")
            raise function_error

        # STEP 2: Function succeeded, now try to add tracing
        try:
            trace_id = request_context.get()
            start_time = datetime.now()
            end_time = datetime.now()
            response_time = end_time - start_time

            # Parse operation data based on operation type
            if self.operation_type == "write":
                operation_data = self.provider.parse_write_response(result)
            else:  # read
                operation_data = self.provider.parse_read_response(result)

            # Calculate cost
            price_data = self.provider.calculate_cost(
                operation_data, operation_data["operation_type"]
            )

            # Create timezone
            ist = timezone(timedelta(hours=5, minutes=30))

            # Get mapped parameter values
            query = self.get_mapped_param_value(kwargs, "query")
            namespace = self.get_mapped_param_value(kwargs, "namespace") or ""
            index_host = self.get_mapped_param_value(kwargs, "index_host") or ""

            # Extract response data
            if self.operation_type == "write":
                vectors_count = result.get("upsertedCount", 0)
                vectors = self.get_mapped_param_value(
                    kwargs, "vectors"
                ) or kwargs.get("vectors", [])
                total_vectors = len(vectors)
                response_data = result
                operation_details = {
                    "index_host": index_host,
                    "namespace": namespace,
                    "upserted_vectors": vectors_count,
                    "total_vectors": total_vectors,
                }
            elif self.operation_type == "read":
                max_results = (
                    self.get_mapped_param_value(kwargs, "max_results")
                    or self.get_mapped_param_value(kwargs, "top_k")
                    or 0
                )
                response_data = result.get("matches", result.get("results", []))
                operation_details = {
                    "index_host": index_host,
                    "namespace": namespace,
                    "top_k": max_results,
                }
            else:
                response_data = result
                operation_details = {}

            span_data = {
                "service_provider": self.provider.name,
                "operation_type": self.operation_type,
                "response": response_data or "",
                "operation_details": operation_details,
                "units": operation_data["units"],
                "price": price_data["price"],
                "query": query,
                "start_time": start_time,
                "end_time": end_time,
                "response_time": response_time,
                "timestamp": datetime.now(ist).strftime("%Y-%m-%d %H:%M:%S"),
            }

            # Build complete metadata
            complete_metadata = {
                **operation_details,
                "operation_type": self.operation_type,
                "provider": self.provider.name,
                "cost": price_data["price"],
                "read_units": operation_data["units"],
            }

            # Filter metadata if metadata_config is provided
            filtered_metadata = self.filter_metadata(complete_metadata)

            await _LangfuseService.create_span_for_vectorDB(
                trace_id=trace_id,
                span_data=span_data,
                name=f"{self.provider.name.capitalize()} {self.operation_type.capitalize()}",
                filtered_metadata=filtered_metadata,
            )

        except Exception as tracing_error:
            log_error(
                tracing_error,
                "Tracing failed - function succeeded, returning result despite tracing failure",
            )

        return result



"""
Concrete vector database provider implementations.
"""

from typing import Any, Dict, List

from ..base.provider import VectorDBProvider


class PineconeVectorDBProvider(VectorDBProvider):
    """Pinecone vector database provider implementation."""

    def __init__(self):
        super().__init__(
            name="pinecone", response_extractor=self._extract_pinecone_response
        )

    def parse_write_response(
        self, response_data: Dict[str, Any]
    ) -> Dict[str, int]:
        """Parse write operation response from Pinecone."""
        return {
            "operation_type": "write",
            "units": response_data.get("upsertedCount", 0),
        }

    def parse_read_response(
        self, response_data: Dict[str, Any]
    ) -> Dict[str, int]:
        """Parse read operation response from Pinecone."""
        usage = response_data.get("usage", {})
        return {"operation_type": "read", "units": usage.get("read_units", 0)}

    def calculate_cost(
        self, operation_data: Dict[str, Any], operation_type: str
    ) -> Dict[str, float]:
        """Calculate cost for Pinecone operations."""
        pricing = {
            "read": 16.0,  # $16 per million read units
            "write": 4.0,  # $4 per million write units
        }

        units = operation_data.get("units", 0)
        price = (units / 1000000) * pricing.get(operation_type, 0.0)

        return {"units": units, "price": price}

    def _extract_pinecone_response(
        self, data: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Extract response data from Pinecone."""
        return [
            {
                "score": match["score"],
                "text": match.get("metadata", {}).get("text", ""),
                "namespace": data.get("namespace", ""),
            }
            for match in data.get("matches", [])
        ]



from typing import Optional

from .langfuse_service import _LangfuseService


class LangfuseInitializer:
    @classmethod
    def initialize(
        cls,
        langfuse_public_key: str,
        langfuse_secret_key: str,
        langfuse_host: str,
        release: str,
        environment: str,
    ) -> None:
        """Initialize the Langfuse client singleton."""
        _LangfuseService.initialize(
            langfuse_public_key=langfuse_public_key,
            langfuse_secret_key=langfuse_secret_key,
            langfuse_host=langfuse_host,
            release=release,
            environment=environment,
        )

    @classmethod
    def get_instance(cls) -> Optional[_LangfuseService]:
        """Get the Langfuse client instance."""
        return _LangfuseService.get_instance()

    @classmethod
    def close(cls) -> None:
        """Close the Langfuse client instance."""
        _LangfuseService.close()



import copy
import json
import time
import uuid
from typing import Any, Dict, List, Optional

from fastapi import Request
from fastapi.responses import StreamingResponse

from ..config.context_util import (
    request_context,
    request_metadata_context,
    tracer_context,
)
from ..config.langfuse_init import LangfuseInitializer


def _preprocess_langfuse_input(input_data: Any) -> Any:
    """
    Preprocess input data to prevent Langfuse's automatic transformation of "messages" keys.

    Args:
        input_data: The original input data

    Returns:
        Processed input data with "messages" keys renamed to avoid automatic parsing
    """
    if input_data is None:
        return input_data

    # Create a deep copy to avoid modifying the original data
    processed_input = copy.deepcopy(input_data)

    def _rename_messages_keys(obj: Any) -> Any:
        """Recursively rename 'messages' keys to 'raw_messages' to prevent auto-parsing"""
        if isinstance(obj, dict):
            new_dict = {}
            for key, value in obj.items():
                if key == "messages":
                    # Rename to prevent Langfuse automatic parsing
                    new_dict["raw_messages"] = _rename_messages_keys(value)
                    # Add a flag to indicate this was transformed
                    new_dict["_langfuse_messages_renamed"] = True
                else:
                    new_dict[key] = _rename_messages_keys(value)
            return new_dict
        elif isinstance(obj, list):
            return [_rename_messages_keys(item) for item in obj]
        else:
            return obj

    return _rename_messages_keys(processed_input)


def _should_trace_route(
    route_path: str,
    include_routes: Optional[List[str]] = None,
    exclude_routes: Optional[List[str]] = None,
) -> bool:
    """
    Determine if a route should be traced based on include/exclude lists.

    Args:
        route_path: The route path to check
        include_routes: List of routes to include for tracing (if provided, only these routes will be traced)
        exclude_routes: List of routes to exclude from tracing

    Returns:
        bool: True if route should be traced, False otherwise

    Priority:
        1. If include_routes is provided, only trace routes in that list
        2. If exclude_routes is provided, trace all routes except those in that list
        3. If both are provided, include_routes takes priority
        4. If neither is provided, trace all routes
    """
    # If include_routes is provided, only trace routes in that list
    if include_routes is not None:
        return route_path in include_routes

    # If exclude_routes is provided, exclude those routes
    if exclude_routes is not None:
        return route_path not in exclude_routes

    # Default: trace all routes
    return True


async def unified_middleware(
    request: Request,
    call_next,
    metadata: Optional[Dict[str, Any]] = None,
    route_mapping: Optional[Dict[str, str]] = None,
    include_routes: Optional[List[str]] = None,
    exclude_routes: Optional[List[str]] = None,
    tag_mapping: Optional[Dict[str, List[str]]] = None,
    input: Optional[Any] = None,
):
    """
    Unified middleware for LLM observability with optional route name mapping, selective tracing, tags support, and input capture.

    Args:
        request: FastAPI request object
        call_next: Next middleware in chain
        metadata: Optional metadata for the trace
        route_mapping: Optional mapping of routes to custom names
                   Example: {"/auth/login": "Authentication with Login", "/user": "User Creation"}
        include_routes: Optional list of routes to include for tracing. If provided, only these routes will be traced.
                       Example: ["/api/chat", "/api/embeddings"]
        exclude_routes: Optional list of routes to exclude from tracing.
                       Example: ["/health", "/metrics", "/docs"]
        tag_mapping: Optional mapping of routes to lists of tags for categorization and filtering
                  Example: {"/api/chat": ["production", "llm"], "/api/embeddings": ["production", "embedding"]}
        input: Optional input data for the trace. Can be any JSON object (e.g., request.body, parsed request data)

    Note:
        - If both include_routes and exclude_routes are provided, include_routes takes priority
        - If a route is not traced, decorators will gracefully skip tracing operations
        - Routes not traced will still process normally but won't create Langfuse traces
        - Tags are used to categorize traces and can be filtered in the Langfuse UI
        - Input data helps track what data was provided to the trace for debugging and analysis
        - Input data containing "messages" keys will be preprocessed to prevent automatic Langfuse transformation
    """
    # Set request context
    session_id = request.headers.get("X-Request-ID", str(uuid.uuid4()))
    request_token = None
    metadata_token = None
    trace_token = None

    # Get the route path from metadata or request
    route_path = metadata.get("apiEndpoint") if metadata else request.url.path

    # Check if this route should be traced
    should_trace = _should_trace_route(
        route_path, include_routes, exclude_routes
    )

    if not should_trace:
        # Skip tracing for this route - just process the request normally
        try:
            response = await call_next(request)

            # Still set basic response headers for consistency
            if response is not None:
                response.headers["X-Request-ID"] = session_id
                response.headers["X-Trace-Status"] = "skipped"

            return response
        except Exception as e:
            print("=" * 40, ">", "Error in non-traced request:", e)
            # return StreamingResponse(
            #     iter(
            #         [json.dumps({"error": f"Internal server error - {str(e)}"})]
            #     ),
            #     status_code=500,
            #     headers={
            #         "X-Request-ID": session_id,
            #         "X-Trace-Status": "skipped",
            #     },
            # )

    # Proceed with tracing for included routes
    trace_id = str(uuid.uuid4())
    langfuse_client = LangfuseInitializer.get_instance().langfuse_client

    if request_context.get() is None:
        request_token = request_context.set(trace_id)

    # Set metadata if provided
    if metadata is not None:
        metadata_token = request_metadata_context.set(metadata)

    # Determine trace name using route_mapping if provided
    if route_mapping and route_path in route_mapping:
        trace_name = route_mapping[route_path]
    else:
        trace_name = route_path if route_path else trace_id

    # Get tags for the current route
    route_tags = tag_mapping.get(route_path, []) if tag_mapping else []

    # Preprocess input to prevent automatic Langfuse transformation
    processed_input = _preprocess_langfuse_input(input)

    if trace_id:
        trace = langfuse_client.trace(
            id=trace_id,
            name=trace_name,
            input=processed_input,  # Use preprocessed input
            session_id=session_id,
            metadata=metadata,
            user_id=metadata.get("user") if metadata else None,
            tags=route_tags,
        )

    trace_token = tracer_context.set(trace)

    request.state.langfuse_context = {
        **(metadata or {}),
        "trace_id": trace_id,
    }

    try:
        # Process the request
        response = await call_next(request)
        end_time = time.time()

        # Ensure we have a valid response before modifying headers
        if response is not None:

            # Set response headers
            response.headers["X-Request-ID"] = session_id
            response.headers["X-Trace-ID"] = trace_id
            response.headers["X-Trace-Status"] = "traced"

            return response
        else:
            print(
                "=" * 40,
                ">",
                "No response returned from call_next from the package",
            )
            # raise RuntimeError("No response returned from call_next")
            return StreamingResponse(
                iter(
                    [
                        json.dumps(
                            {
                                "error": "Internal server error - No response from service"
                            }
                        )
                    ]
                ),
                status_code=500,
                headers={
                    "X-Request-ID": session_id,
                    "X-Trace-ID": trace_id,
                    "X-Trace-Status": "traced",
                },
            )
    except Exception as e:
        print("=" * 40, ">", "Error in the middleware from the package", str(e))
        # return StreamingResponse(
        #     iter([json.dumps({"error": f"Internal server error - {str(e)}"})]),
        #     status_code=500,
        #     headers={
        #         "X-Request-ID": session_id,
        #         "X-Trace-ID": trace_id,
        #         "X-Trace-Status": "traced",
        #     },
        # )
    finally:
        # Clean up context tokens
        if request_token is not None:
            request_context.reset(request_token)
        if metadata_token is not None:
            request_metadata_context.reset(metadata_token)
        if trace_token is not None:
            tracer_context.reset(trace_token)



"""
Concrete LLM provider implementations.
"""

import time
from typing import Any, Dict

from ..base.provider import LLMProvider
from ..utils.token_costs import get_token_costs
from ..utils.tool_call_parser import extract_anthropic_response_with_tools


class OpenAIProvider(LLMProvider):
    """OpenAI LLM provider implementation."""

    def __init__(self):
        super().__init__(
            name="openai", response_extractor=self._extract_openai_response
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from OpenAI response."""
        # Handle None response (e.g., when request is cancelled)
        if response_data is None:
            return {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0,
            }
        
        usage = response_data.get("usage", {})
        return {
            "prompt_tokens": usage.get("prompt_tokens", 0),
            "completion_tokens": usage.get("completion_tokens", 0),
            "total_tokens": usage.get("total_tokens", 0),
        }

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for OpenAI model."""
        try:
            model_pricing = get_token_costs(model_name, self.name)
            prompt_tokens = tokens_data.get("prompt_tokens", 0)
            completion_tokens = tokens_data.get("completion_tokens", 0)

            input_price = prompt_tokens * model_pricing["input_cost_per_token"]
            output_price = (
                completion_tokens * model_pricing["output_cost_per_token"]
            )
            total_price = input_price + output_price

            return {
                "input": input_price,
                "output": output_price,
                "total": total_price,
            }
        except ValueError as e:
            raise e

    def get_completion_start_time(self, response_data: Dict[str, Any]) -> float:
        """Extract completion start time from OpenAI response."""
        return response_data.get("created", time.time())

    def _extract_openai_response(self, data: Dict[str, Any]) -> str:
        """Extract response content from OpenAI response."""
        if "choices" in data and data["choices"]:
            return data["choices"][0]["message"]["content"]
        return ""


class AnthropicProvider(LLMProvider):
    """Anthropic LLM provider implementation."""

    def __init__(self):
        super().__init__(
            name="anthropic",
            response_extractor=extract_anthropic_response_with_tools,
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Anthropic response."""
        # Handle None response (e.g., when request is cancelled)
        if response_data is None:
            return {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0,
                "cache_creation_input_tokens": 0,
                "cache_read_input_tokens": 0,
            }
        
        usage = response_data.get("usage", {})
        input_tokens = usage.get("input_tokens", 0)
        output_tokens = usage.get("output_tokens", 0)
        cache_creation_input_tokens = usage.get(
            "cache_creation_input_tokens", 0
        )
        cache_read_input_tokens = usage.get("cache_read_input_tokens", 0)

        return {
            "prompt_tokens": input_tokens,
            "completion_tokens": output_tokens,
            "total_tokens": input_tokens + output_tokens,
            "cache_creation_input_tokens": cache_creation_input_tokens,
            "cache_read_input_tokens": cache_read_input_tokens,
        }

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Anthropic model."""
        try:
            model_pricing = get_token_costs(model_name, self.name)
            prompt_tokens = tokens_data.get("prompt_tokens", 0)
            completion_tokens = tokens_data.get("completion_tokens", 0)
            cache_creation_input_tokens = tokens_data.get(
                "cache_creation_input_tokens", 0
            )
            cache_read_input_tokens = tokens_data.get(
                "cache_read_input_tokens", 0
            )

            input_price = prompt_tokens * model_pricing["input_cost_per_token"]
            output_price = (
                completion_tokens * model_pricing["output_cost_per_token"]
            )

            # Calculate cache-related costs
            cache_creation_price = 0.0
            cache_read_price = 0.0

            if (
                cache_creation_input_tokens > 0
                and "cache_creation_input_token_cost" in model_pricing
            ):
                cache_creation_price = (
                    cache_creation_input_tokens
                    * model_pricing["cache_creation_input_token_cost"]
                )

            if (
                cache_read_input_tokens > 0
                and "cache_read_input_token_cost" in model_pricing
            ):
                cache_read_price = (
                    cache_read_input_tokens
                    * model_pricing["cache_read_input_token_cost"]
                )

            total_price = (
                input_price
                + output_price
                + cache_creation_price
                + cache_read_price
            )

            return {
                "input": input_price,
                "output": output_price,
                "cache_creation": cache_creation_price,
                "cache_read": cache_read_price,
                "total": total_price,
            }
        except ValueError as e:
            raise e

    def get_completion_start_time(self, response_data: Dict[str, Any]) -> float:
        """Extract completion start time from Anthropic response."""
        # Handle None response (e.g., when request is cancelled)
        if response_data is None:
            return time.time()
        
        return response_data.get("started_at", time.time())


class GroqProvider(LLMProvider):
    """Groq LLM provider implementation."""

    def __init__(self):
        super().__init__(
            name="groq", response_extractor=self._extract_groq_response
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Groq response."""
        # Handle None response (e.g., when request is cancelled)
        if response_data is None:
            return {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0,
            }
        
        usage = response_data.get("usage", {})
        return {
            "prompt_tokens": usage.get("prompt_tokens", 0),
            "completion_tokens": usage.get("completion_tokens", 0),
            "total_tokens": usage.get("total_tokens", 0),
        }

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Groq model."""
        try:
            model_pricing = get_token_costs(model_name, self.name)
            prompt_tokens = tokens_data.get("prompt_tokens", 0)
            completion_tokens = tokens_data.get("completion_tokens", 0)

            input_price = prompt_tokens * model_pricing["input_cost_per_token"]
            output_price = (
                completion_tokens * model_pricing["output_cost_per_token"]
            )
            total_price = input_price + output_price

            return {
                "input": input_price,
                "output": output_price,
                "total": total_price,
            }
        except ValueError as e:
            raise e

    def get_completion_start_time(self, response_data: Dict[str, Any]) -> float:
        """Extract completion start time from Groq response."""
        return response_data.get("created", time.time())

    def _extract_groq_response(self, data: Dict[str, Any]) -> str:
        """Extract response content from Groq response."""
        if "choices" in data and data["choices"]:
            return data["choices"][0]["message"]["content"]
        return ""


class GeminiProvider(LLMProvider):
    """Gemini LLM provider implementation."""

    def __init__(self):
        super().__init__(
            name="gemini", response_extractor=self._extract_gemini_response
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Gemini response with detailed modality breakdown."""
        # Handle None response (e.g., when request is cancelled)
        if response_data is None:
            return {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0,
            }
        
        usage = response_data.get("usageMetadata", {})

        # Basic token counts
        tokens_data = {
            "prompt_tokens": usage.get("promptTokenCount", 0),
            "completion_tokens": usage.get("candidatesTokenCount", 0),
            "total_tokens": usage.get("totalTokenCount", 0),
        }

        # Extract detailed token breakdown by modality
        prompt_details = usage.get("promptTokensDetails", [])
        candidates_details = usage.get("candidatesTokensDetails", [])

        # Parse prompt tokens by modality
        for detail in prompt_details:
            if isinstance(detail, dict):
                modality = detail.get("modality", "").lower()
                token_count = detail.get("tokenCount", 0)
                if modality:
                    tokens_data[f"prompt_{modality}_tokens"] = token_count

        # Parse completion tokens by modality
        for detail in candidates_details:
            if isinstance(detail, dict):
                modality = detail.get("modality", "").lower()
                token_count = detail.get("tokenCount", 0)
                if modality:
                    tokens_data[f"completion_{modality}_tokens"] = token_count

        return tokens_data

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Gemini model."""
        try:
            model_pricing = get_token_costs(model_name, self.name)
            prompt_tokens = tokens_data.get("prompt_tokens", 0)
            completion_tokens = tokens_data.get("completion_tokens", 0)

            input_price = prompt_tokens * model_pricing["input_cost_per_token"]
            output_price = (
                completion_tokens * model_pricing["output_cost_per_token"]
            )
            total_price = input_price + output_price

            return {
                "input": input_price,
                "output": output_price,
                "total": total_price,
            }
        except ValueError as e:
            raise e

    def get_completion_start_time(self, response_data: Dict[str, Any]) -> float:
        """Extract completion start time from Gemini response."""
        create_time = response_data.get("createTime")
        if create_time:
            # Convert ISO timestamp to Unix timestamp
            from datetime import datetime

            try:
                dt = datetime.fromisoformat(create_time.replace("Z", "+00:00"))
                return dt.timestamp()
            except ValueError:
                pass
        return time.time()

    def _extract_gemini_response(self, data: Dict[str, Any]) -> str:
        """Extract response content from Gemini response."""
        candidates = data.get("candidates", [])
        if not candidates:
            return ""

        candidate = candidates[0]
        content = candidate.get("content", {})
        parts = content.get("parts", [])

        response_text = ""
        for part in parts:
            if "text" in part:
                response_text += part["text"]

        return response_text



"""
OOP reranking tracer implementation.
"""

from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, List, Optional

from ..base.tracer import BaseTracer
from ..config.context_util import request_context
from ..config.langfuse_service import _LangfuseService
from ..utils.tracer_utils import log_error


class RerankingTracer(BaseTracer):
    """OOP implementation of reranking tracing."""

    def __init__(
        self,
        provider,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        super().__init__(provider, variable_mapping, metadata_config)

    async def trace(self, func: Callable, *args, **kwargs) -> Any:
        """Execute reranking tracing for the function call."""
        if not self.should_trace():
            return await func(*args, **kwargs)

        # STEP 1: Always execute the function first
        try:
            result = await func(*args, **kwargs)
        except Exception as function_error:
            log_error(function_error, "Original function failed")
            raise function_error

        # STEP 2: Function succeeded, now try to add tracing
        try:
            trace_id = request_context.get()
            start_time = datetime.now()
            end_time = datetime.now()
            response_time = end_time - start_time

            # Parse tokens and calculate cost
            tokens_data = self.provider.parse_tokens(result)

            # Get mapped parameter values
            model_name = self.get_mapped_param_value(kwargs, "model_name")
            query = self.get_mapped_param_value(kwargs, "query")
            documents = self.get_mapped_param_value(kwargs, "documents")
            top_n = self.get_mapped_param_value(kwargs, "top_n")

            # Ensure documents is a list
            if not documents:
                documents = []
            elif not isinstance(documents, list):
                documents = [documents]

            document_count = len(documents)

            # Calculate price
            price_details = 0.0
            if tokens_data and model_name:
                try:
                    cost_data = self.provider.calculate_cost(
                        tokens_data, model_name
                    )
                    price_details = cost_data.get("total", 0.0)
                except Exception:
                    price_details = 0.0

            # Extract reranking results
            rerank_results = self.provider.extract_results(result)

            # Create timezone
            ist = timezone(timedelta(hours=5, minutes=30))

            span_data = {
                "service_provider": self.provider.name,
                "model_name": model_name,
                "query": query,
                "documents": documents,
                "document_count": document_count,
                "rerank_results": rerank_results,
                "tokens": tokens_data,
                "price": price_details,
                "top_n": top_n,
                "start_time": start_time,
                "end_time": end_time,
                "response_time": response_time,
                "timestamp": datetime.now(ist).strftime("%Y-%m-%d %H:%M:%S"),
            }

            # Build complete metadata
            complete_metadata = {
                "provider": self.provider.name,
                "model_name": model_name,
                "output_count": document_count,
                "cost": price_details,
                "token usage": tokens_data.get("rerank_units", 0),
                "timestamp": datetime.now(ist).strftime("%Y-%m-%d %H:%M:%S"),
                "top_n": top_n,
            }

            # Filter metadata if metadata_config is provided
            filtered_metadata = self.filter_metadata(complete_metadata)

            await _LangfuseService.create_span_for_reranking(
                trace_id=trace_id,
                span_data=span_data,
                name=f"{self.provider.name.capitalize()} Reranking",
                filtered_metadata=filtered_metadata,
            )

        except Exception as tracing_error:
            log_error(
                tracing_error,
                "Tracing failed - function succeeded, returning result despite tracing failure",
            )

        return result



"""
Base tracer classes and tracer management.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

from ..config.context_util import request_context
from ..utils.tracer_utils import filter_metadata, get_mapped_param_value


class BaseTracer(ABC):
    """Abstract base class for all tracers."""

    def __init__(
        self,
        provider,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        self.provider = provider
        self.variable_mapping = variable_mapping or {}
        self.metadata_config = metadata_config

    @abstractmethod
    async def trace(self, func: Callable, *args, **kwargs) -> Any:
        """Execute tracing for the function call."""

    def get_mapped_param_value(
        self, params: Dict[str, Any], expected_key: str
    ) -> Any:
        """Get parameter value using variable mapping."""
        return get_mapped_param_value(
            params, expected_key, self.variable_mapping
        )

    def filter_metadata(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        """Filter metadata based on metadata_config."""
        return filter_metadata(metadata, self.metadata_config)

    def should_trace(self) -> bool:
        """Check if tracing should be performed."""
        trace_id = request_context.get()
        return trace_id is not None


class TracerManager:
    """Manages all tracers and providers."""

    def __init__(self):
        self._llm_providers = {}
        self._embedding_providers = {}
        self._vectordb_providers = {}
        self._reranking_providers = {}

    def register_llm_provider(self, name: str, provider):
        """Register an LLM provider."""
        self._llm_providers[name] = provider

    def register_embedding_provider(self, name: str, provider):
        """Register an embedding provider."""
        self._embedding_providers[name] = provider

    def register_vectordb_provider(self, name: str, provider):
        """Register a vector DB provider."""
        self._vectordb_providers[name] = provider

    def register_reranking_provider(self, name: str, provider):
        """Register a reranking provider."""
        self._reranking_providers[name] = provider

    def get_llm_provider(self, name: str):
        """Get LLM provider by name."""
        return self._llm_providers.get(name)

    def get_embedding_provider(self, name: str):
        """Get embedding provider by name."""
        return self._embedding_providers.get(name)

    def get_vectordb_provider(self, name: str):
        """Get vector DB provider by name."""
        return self._vectordb_providers.get(name)

    def get_reranking_provider(self, name: str):
        """Get reranking provider by name."""
        return self._reranking_providers.get(name)


# Global tracer manager instance
tracer_manager = TracerManager()



from langfuse.client import StatefulTraceClient

from .context_util import request_context, tracer_context


def get_request_id() -> str:
    return request_context.get()


def get_tracer() -> StatefulTraceClient:
    return tracer_context.get()



from typing import Any, Dict, Optional

from langfuse import Langfuse

from .context_util import tracer_context


class LangfuseClient:
    def __init__(
        self,
        langfuse_public_key: str,
        langfuse_secret_key: str,
        langfuse_host: str,
        release: str,
        environment: str,
    ):
        self.langfuse_public_key = langfuse_public_key
        self.langfuse_secret_key = langfuse_secret_key
        self.langfuse_host = langfuse_host
        self.release = release
        self.environment = environment
        self.langfuse_client = None

    def initialize_langfuse_client(self):
        self.langfuse_client = Langfuse(
            public_key=self.langfuse_public_key,
            secret_key=self.langfuse_secret_key,
            host=self.langfuse_host,
            release=self.release,
            environment=self.environment,
        )

    def close_langfuse_client(self):
        pass


class _LangfuseService:
    _instance = None

    @classmethod
    def initialize(
        cls,
        langfuse_public_key: str,
        langfuse_secret_key: str,
        langfuse_host: str,
        release: str,
        environment: str,
    ) -> None:
        """Initialize the Langfuse client singleton."""
        if cls._instance is None:
            cls._instance = LangfuseClient(
                langfuse_public_key=langfuse_public_key,
                langfuse_secret_key=langfuse_secret_key,
                langfuse_host=langfuse_host,
                release=release,
                environment=environment,
            )
            cls._instance.initialize_langfuse_client()

    @classmethod
    def get_instance(cls) -> Optional[LangfuseClient]:
        """Get the Langfuse client instance."""
        return cls._instance

    @classmethod
    def close(cls) -> None:
        """Close the Langfuse client instance."""
        if cls._instance is not None:
            cls._instance.close_langfuse_client()
            cls._instance = None

    @staticmethod
    async def create_generation_for_LLM(
        trace_id: str,
        generation_data: Dict[str, Any],
        name: str,
        filtered_metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[str]:

        try:
            trace = tracer_context.get()

            trace.update(
                output=generation_data["output"],
            )

            langfuse_client = _LangfuseService.get_instance().langfuse_client
            generation_object = langfuse_client.generation(
                trace_id=trace_id,
                name=name,
            )

            # Build usage details with cache tokens
            usage_details = {
                "input": generation_data["usage"]["prompt_tokens"],
                "output": generation_data["usage"]["completion_tokens"],
                # "total_token": generation_data["usage"]["total_tokens"],
            }

            # Add cache tokens if present
            if (
                generation_data["usage"].get("cache_creation_input_tokens", 0)
                > 0
            ):
                usage_details["cache_creation_input_tokens"] = generation_data[
                    "usage"
                ]["cache_creation_input_tokens"]

            if generation_data["usage"].get("cache_read_input_tokens", 0) > 0:
                usage_details["cache_read_input_tokens"] = generation_data[
                    "usage"
                ]["cache_read_input_tokens"]

            # Build cost details with cache costs
            base_input_cost = generation_data["cost_details"]["input"]
            base_output_cost = generation_data["cost_details"]["output"]
            cache_creation_cost = generation_data["cost_details"].get(
                "cache_creation", 0.0
            )
            cache_read_cost = generation_data["cost_details"].get(
                "cache_read", 0.0
            )

            # Calculate total cost including cache costs
            total_cost_with_cache = (
                base_input_cost
                + base_output_cost
                + cache_creation_cost
                + cache_read_cost
            )

            cost_details = {
                "input": base_input_cost,
                "output": base_output_cost,
                "total": total_cost_with_cache,  # ✅ FIXED: Include cache costs in total
            }

            # Add cache costs if present
            if cache_creation_cost > 0.0:
                cost_details["cache_creation"] = cache_creation_cost

            if cache_read_cost > 0.0:
                cost_details["cache_read"] = cache_read_cost

            generation_object.end(
                model=generation_data["model_name"],
                input=generation_data["input"],
                output=generation_data["output"],
                usage_details=usage_details,
                cost_details=cost_details,
                metadata=(
                    filtered_metadata
                    if filtered_metadata is not None
                    else {
                        "model": generation_data["model_name"],
                        "provider": generation_data["service_provider"],
                        "maxTokens": generation_data["max_tokens"],
                        "temperature": generation_data["temperature"],
                        "tool": generation_data["tool_names"],
                        "timeTaken": (
                            generation_data["end_time"]
                            - generation_data["start_time"]
                        ).total_seconds(),
                        "inputTokens": generation_data["usage"][
                            "prompt_tokens"
                        ],
                        "outputTokens": generation_data["usage"][
                            "completion_tokens"
                        ],
                        "inputCost": base_input_cost,
                        "outputCost": base_output_cost,
                        "totalCost": total_cost_with_cache,  # ✅ FIXED: Use total cost including cache costs
                        # Cache-specific costs in metadata for detailed tracking
                        **(
                            {"cacheCreationCost": cache_creation_cost}
                            if cache_creation_cost > 0.0
                            else {}
                        ),
                        **(
                            {"cacheReadCost": cache_read_cost}
                            if cache_read_cost > 0.0
                            else {}
                        ),
                        **generation_data.get("tool_call_metadata", {}),
                    }
                ),
            )
            return generation_object.id
        except Exception as e:
            return None

    @staticmethod
    async def create_span_for_vectorDB(
        trace_id: str,
        span_data: Dict[str, Any],
        name: str,
        filtered_metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[str]:
        try:

            langfuse_client = _LangfuseService.get_instance().langfuse_client
            span_object = langfuse_client.span(
                trace_id=trace_id,
                name=name,
            )

            span_object.end(
                input=span_data["query"],
                output=span_data["response"],
                metadata=(
                    filtered_metadata
                    if filtered_metadata is not None
                    else {
                        **span_data["operation_details"],
                        "operation_type": span_data["operation_type"],
                        "provider": span_data["service_provider"],
                        "cost": span_data["price"],
                        "read_units": span_data["units"],
                    }
                ),
            )

            return span_object.id
        except Exception as e:
            return None

    @staticmethod
    async def create_span_for_embedding(
        trace_id: str,
        span_data: Dict[str, Any],
        name: str,
        filtered_metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[str]:

        try:
            trace = tracer_context.get()

            trace.update(
                input=span_data["input"],
            )

            langfuse_client = _LangfuseService.get_instance().langfuse_client
            span_object = langfuse_client.span(
                trace_id=trace_id,
                name=name,
            )

            span_object.end(
                input=span_data["input"],
                metadata=(
                    filtered_metadata
                    if filtered_metadata is not None
                    else {
                        "provider": span_data["service_provider"],
                        "model_name": span_data["model_name"],
                        "input count": span_data["input_count"],
                        "cost": span_data["price"]["total"],
                        "token usage": span_data["tokens"],
                        "price": span_data["price"],
                        "embedding_dimensions": span_data[
                            "embedding_dimensions"
                        ],
                        "timestamp": span_data["timestamp"],
                    }
                ),
            )

            return span_object.id
        except Exception as e:
            return None

    @staticmethod
    async def create_span_for_reranking(
        trace_id: str,
        span_data: Dict[str, Any],
        name: str,
        filtered_metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[str]:
        try:
            langfuse_client = _LangfuseService.get_instance().langfuse_client
            span_object = langfuse_client.span(
                trace_id=trace_id,
                name=name,
            )

            span_object.end(
                input={
                    "query": span_data["query"],
                    "documents": span_data["documents"],
                },
                output=span_data["rerank_results"],
                metadata=(
                    filtered_metadata
                    if filtered_metadata is not None
                    else {
                        "provider": span_data["service_provider"],
                        "model_name": span_data["model_name"],
                        "output_count": span_data["document_count"],
                        "cost": span_data["price"],
                        "token usage": span_data["tokens"]["rerank_units"],
                        "timestamp": span_data["timestamp"],
                        "top_n": span_data["top_n"],
                    }
                ),
            )

            return span_object.id
        except Exception as e:
            return None

    @staticmethod
    async def create_span_for_general(
        trace_id: str,
        span_data: Dict[str, Any],
        name: str,
        filtered_metadata: Optional[Dict[str, Any]] = None,
        parent_observation_id: Optional[str] = None,
    ) -> Optional[str]:
        """
        Create a general-purpose span for any function.

        Args:
            trace_id: The trace ID
            span_data: Dict containing input, output, start_time, end_time, function_name
            name: Name of the span
            filtered_metadata: Optional filtered metadata
            parent_observation_id: Optional parent observation ID for nesting

        Returns:
            The span ID if successful, None otherwise
        """
        try:
            langfuse_client = _LangfuseService.get_instance().langfuse_client

            # Create span with optional parent
            span_kwargs = {
                "trace_id": trace_id,
                "name": name,
                "start_time": span_data["start_time"],
            }

            if parent_observation_id:
                span_kwargs["parent_observation_id"] = parent_observation_id

            span_object = langfuse_client.span(**span_kwargs)

            span_object.end(
                end_time=span_data["end_time"],
                input=span_data["input"],
                output=span_data["output"],
                metadata=(
                    filtered_metadata if filtered_metadata is not None else {}
                ),
            )

            return span_object.id
        except Exception as e:
            return None



"""
General-purpose tracer implementation for any function.
"""

import inspect
from contextvars import ContextVar
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional

from ..base.tracer import BaseTracer
from ..config.context_util import tracer_context
from ..config.langfuse_service import _LangfuseService

# Context variable to store current span ID for nesting
current_span_context: ContextVar[Optional[str]] = ContextVar(
    "current_span_context", default=None
)


class GeneralTracer(BaseTracer):
    """General-purpose tracer for any function."""

    def __init__(
        self,
        span_name: Optional[str] = None,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        # Use a dummy provider since this is a general tracer
        class DummyProvider:
            name = "general"

        super().__init__(DummyProvider(), variable_mapping, metadata_config)
        self.span_name = span_name

    async def trace(self, func: Callable, *args, **kwargs) -> Any:
        """Execute general tracing for the function call."""
        if not self.should_trace():
            return await func(*args, **kwargs)

        # Get trace context
        trace = tracer_context.get()
        if not trace:
            return await func(*args, **kwargs)

        start_time = datetime.now()

        # Determine span name
        span_name = self.span_name or func.__name__.replace("_", " ").title()

        # Get current parent span for nesting
        parent_span_id = current_span_context.get()

        # Prepare input data
        input_data = self._prepare_input_data(func, args, kwargs)

        # Clear any previously captured result
        clear_captured_result()

        span_token = None
        actual_span_id = None

        # First, create a placeholder span to get an ID for child spans
        # We'll use a two-phase approach: create span, execute function, then update span
        span_kwargs = {
            "trace_id": trace.id,
            "name": span_name,
            "start_time": start_time,
        }

        if parent_span_id:
            span_kwargs["parent_observation_id"] = parent_span_id

        # Create the initial span to get a span ID
        langfuse_client = _LangfuseService.get_instance().langfuse_client
        span_object = langfuse_client.span(**span_kwargs)
        actual_span_id = span_object.id

        # Set this span ID in context BEFORE executing the function so child spans can reference it
        span_token = current_span_context.set(actual_span_id)

        try:
            # Execute the function (now child spans will have the correct parent ID)
            result = await func(*args, **kwargs)
            end_time = datetime.now()

            # Check if a result was captured (for functions without return values)
            captured_result = get_captured_result()
            output_data = self._prepare_output_data(
                captured_result if captured_result is not None else result
            )

            # Build complete metadata
            complete_metadata = {
                "functionName": func.__name__,
                "timeTaken": (end_time - start_time).total_seconds(),
                "hasReturn": result is not None,
                "hasCapturedResult": captured_result is not None,
                "argumentCount": len(args) + len(kwargs),
                "isAsync": inspect.iscoroutinefunction(func),
                "module": (
                    func.__module__
                    if hasattr(func, "__module__")
                    else "unknown"
                ),
            }

            # Filter metadata if metadata_config is provided
            filtered_metadata = self.filter_metadata(complete_metadata)

            # Update the span with the final input, output, and metadata
            span_object.end(
                end_time=end_time,
                input=input_data,
                output=output_data,
                metadata=(
                    filtered_metadata if filtered_metadata is not None else {}
                ),
            )

            return result

        except Exception as e:
            end_time = datetime.now()

            # Build error metadata
            complete_metadata = {
                "functionName": func.__name__,
                "timeTaken": (end_time - start_time).total_seconds(),
                "hasError": True,
                "errorType": type(e).__name__,
                "argumentCount": len(args) + len(kwargs),
                "isAsync": inspect.iscoroutinefunction(func),
                "module": (
                    func.__module__
                    if hasattr(func, "__module__")
                    else "unknown"
                ),
            }

            # Filter metadata if metadata_config is provided
            filtered_metadata = self.filter_metadata(complete_metadata)

            # Update the span with error information
            span_object.end(
                end_time=end_time,
                input=input_data,
                output={"error": str(e), "error_type": type(e).__name__},
                metadata=(
                    filtered_metadata if filtered_metadata is not None else {}
                ),
            )

            raise e

        finally:
            # Reset the span context
            if span_token:
                current_span_context.reset(span_token)

    def _prepare_input_data(
        self, func: Callable, args: tuple, kwargs: dict
    ) -> Dict[str, Any]:
        """Prepare input data for the span."""
        try:
            # Get function signature
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            # Convert to dictionary, handling special cases
            input_dict = {}
            for param_name, value in bound_args.arguments.items():
                if param_name == "self":
                    input_dict[param_name] = str(type(value).__name__)
                else:
                    input_dict[param_name] = self._serialize_value(value)

            return input_dict
        except Exception:
            # Fallback to simple args/kwargs structure
            return {
                "args": [self._serialize_value(arg) for arg in args],
                "kwargs": {
                    k: self._serialize_value(v) for k, v in kwargs.items()
                },
            }

    def _prepare_output_data(self, result: Any) -> Any:
        """Prepare output data for the span."""
        return self._serialize_value(result)

    def _serialize_value(self, value: Any) -> Any:
        """Serialize a value for JSON storage, handling special types."""
        try:
            # Handle None
            if value is None:
                return None

            # Handle basic types
            if isinstance(value, (str, int, float, bool)):
                return value

            # Handle lists and tuples
            if isinstance(value, (list, tuple)):
                return [
                    self._serialize_value(item) for item in value[:10]
                ]  # Limit to first 10 items

            # Handle dictionaries
            if isinstance(value, dict):
                return {
                    k: self._serialize_value(v)
                    for k, v in list(value.items())[:10]
                }  # Limit to first 10 items

            # Handle objects with string representation
            str_repr = str(value)
            if len(str_repr) > 1000:  # Limit string length
                return str_repr[:1000] + "..."
            return str_repr

        except Exception:
            return f"<{type(value).__name__} object>"


# Context manager for capturing results in functions that don't return anything
class ResultCapture:
    """Context manager for capturing results in functions without return values."""

    def __init__(self):
        self._result = None
        self._captured = False

    def capture_result(self, result: Any) -> None:
        """Capture a result to be used as span output."""
        self._result = result
        self._captured = True

    def get_result(self) -> Any:
        """Get the captured result."""
        return self._result if self._captured else None

    def has_result(self) -> bool:
        """Check if a result was captured."""
        return self._captured


# Global result capture instance that can be imported
_result_capture = ResultCapture()


def capture_result(result: Any) -> None:
    """Global function to capture results for spans."""
    _result_capture.capture_result(result)


def get_captured_result() -> Any:
    """Get the captured result."""
    return _result_capture.get_result()


def clear_captured_result() -> None:
    """Clear the captured result."""
    global _result_capture
    _result_capture = ResultCapture()



"""
Base interfaces and abstract classes for the observability framework.
"""

from .provider import (
    BaseProvider,
    EmbeddingProvider,
    LLMProvider,
    RerankingProvider,
    VectorDBProvider,
)
from .tracer import BaseTracer, TracerManager

__all__ = [
    "BaseProvider",
    "LLMProvider",
    "EmbeddingProvider",
    "VectorDBProvider",
    "RerankingProvider",
    "BaseTracer",
    "TracerManager",
]



"""
Main service manager for the observability framework.
"""

import functools
from typing import Any, Callable, Dict, List, Optional

from ..base.tracer import tracer_manager
from ..providers import (
    AnthropicProvider,
    CohereEmbeddingProvider,
    CohereRerankingProvider,
    GeminiProvider,
    GroqProvider,
    JinaEmbeddingProvider,
    JinaRerankingProvider,
    OpenAIProvider,
    PineconeEmbeddingProvider,
    PineconeRerankingProvider,
    PineconeVectorDBProvider,
    VoyageAIEmbeddingProvider,
    VoyageRerankingProvider,
)
from ..tracers import (
    EmbeddingTracer,
    GeneralTracer,
    LLMStreamingTracer,
    LLMTracer,
    RerankingTracer,
    VectorDBTracer,
)


class ObservabilityService:
    """Main service manager for AI observability with OOP design."""

    def __init__(self):
        self._setup_providers()

    def _setup_providers(self):
        """Setup all default providers."""
        # Register LLM providers
        tracer_manager.register_llm_provider("openai", OpenAIProvider())
        tracer_manager.register_llm_provider("anthropic", AnthropicProvider())
        tracer_manager.register_llm_provider("groq", GroqProvider())
        tracer_manager.register_llm_provider("gemini", GeminiProvider())

        # Register embedding providers
        tracer_manager.register_embedding_provider(
            "pinecone", PineconeEmbeddingProvider()
        )
        tracer_manager.register_embedding_provider(
            "cohere", CohereEmbeddingProvider()
        )
        tracer_manager.register_embedding_provider(
            "jina", JinaEmbeddingProvider()
        )
        tracer_manager.register_embedding_provider(
            "voyageai", VoyageAIEmbeddingProvider()
        )

        # Register vector DB providers
        tracer_manager.register_vectordb_provider(
            "pinecone", PineconeVectorDBProvider()
        )

        # Register reranking providers
        tracer_manager.register_reranking_provider(
            "pinecone", PineconeRerankingProvider()
        )
        tracer_manager.register_reranking_provider(
            "cohere", CohereRerankingProvider()
        )
        tracer_manager.register_reranking_provider(
            "jina", JinaRerankingProvider()
        )
        tracer_manager.register_reranking_provider(
            "voyage", VoyageRerankingProvider()
        )

    def register_llm_provider(self, name: str, provider):
        """Register a custom LLM provider."""
        tracer_manager.register_llm_provider(name, provider)

    def register_embedding_provider(self, name: str, provider):
        """Register a custom embedding provider."""
        tracer_manager.register_embedding_provider(name, provider)

    def register_vectordb_provider(self, name: str, provider):
        """Register a custom vector DB provider."""
        tracer_manager.register_vectordb_provider(name, provider)

    def register_reranking_provider(self, name: str, provider):
        """Register a custom reranking provider."""
        tracer_manager.register_reranking_provider(name, provider)

    def llm_tracing(
        self,
        provider: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
        is_sdk: bool = False,
    ):
        """
        Decorator for tracing LLM API calls with provider-specific handling.

        Args:
            provider: Name of the LLM provider (e.g., "openai", "anthropic", "groq")
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.
            is_sdk: Whether to treat the function as SDK mode (returns complete response) or standard mode (default: False)
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(self, **params):
                provider_instance = tracer_manager.get_llm_provider(provider)
                if not provider_instance:
                    return await func(self, **params)
                tracer = LLMTracer(
                    provider_instance, variable_mapping, metadata_config, is_sdk
                )
                return await tracer.trace(func, self, **params)

            return wrapper

        return decorator

    def llm_streaming_tracing(
        self,
        provider: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
        is_sdk: bool = False,
    ):
        """
        Decorator for tracing streaming LLM API calls.
        Currently only supports Anthropic provider.

        Args:
            provider: Name of the LLM provider (must be "anthropic")
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.
            is_sdk: Whether to treat the function as SDK mode (returns complete response) or streaming mode (async generator)
        """

        def decorator(func):
            if is_sdk:
                # NEW SDK mode: Return async generator function that handles anthropic_final_message events
                @functools.wraps(func)
                async def wrapper(self, **params):
                    provider_instance = tracer_manager.get_llm_provider(
                        provider
                    )
                    if not provider_instance:
                        async for response_line in func(self, **params):
                            yield response_line
                        return

                    tracer = LLMStreamingTracer(
                        provider_instance,
                        variable_mapping,
                        metadata_config,
                        is_sdk,
                    )
                    async for response_line in tracer.trace(
                        func, self, **params
                    ):
                        yield response_line

                return wrapper
            else:
                # Streaming mode: Return async generator function
                @functools.wraps(func)
                async def wrapper(self, **params):
                    provider_instance = tracer_manager.get_llm_provider(
                        provider
                    )
                    if not provider_instance:
                        async for response_line in func(self, **params):
                            yield response_line
                        return

                    tracer = LLMStreamingTracer(
                        provider_instance,
                        variable_mapping,
                        metadata_config,
                        is_sdk,
                    )
                    async for response_line in tracer.trace(
                        func, self, **params
                    ):
                        yield response_line

                return wrapper

        return decorator

    def embedding_tracing(
        self,
        provider: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        """
        Decorator for tracing embedding API calls with provider-specific handling.

        Args:
            provider: Name of the embedding provider (e.g., "pinecone", "cohere", "jina", "voyageai")
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                provider_instance = tracer_manager.get_embedding_provider(
                    provider
                )
                if not provider_instance:
                    return await func(*args, **kwargs)

                tracer = EmbeddingTracer(
                    provider_instance, variable_mapping, metadata_config
                )
                return await tracer.trace(func, *args, **kwargs)

            return wrapper

        return decorator

    def vectordb_tracing(
        self,
        provider: str,
        operation_type: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        """
        Decorator for tracing Vector DB API calls.

        Args:
            provider: Name of the vector DB provider (e.g., "pinecone")
            operation_type: Type of operation ("read" or "write")
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                provider_instance = tracer_manager.get_vectordb_provider(
                    provider
                )
                if not provider_instance:
                    return await func(*args, **kwargs)

                tracer = VectorDBTracer(
                    provider_instance,
                    operation_type,
                    variable_mapping,
                    metadata_config,
                )
                return await tracer.trace(func, *args, **kwargs)

            return wrapper

        return decorator

    def reranking_tracing(
        self,
        provider: str,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        """
        Decorator for tracing reranking API calls with provider-specific handling.

        Args:
            provider: Name of the reranking provider (e.g., "pinecone", "cohere", "jina", "voyage")
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                provider_instance = tracer_manager.get_reranking_provider(
                    provider
                )
                if not provider_instance:
                    return await func(*args, **kwargs)

                tracer = RerankingTracer(
                    provider_instance, variable_mapping, metadata_config
                )
                return await tracer.trace(func, *args, **kwargs)

            return wrapper

        return decorator

    def general_tracing(
        self,
        name: Optional[str] = None,
        variable_mapping: Optional[Dict[str, str]] = None,
        metadata_config: Optional[List[str]] = None,
    ):
        """
        General-purpose decorator for tracing any function.

        Args:
            name: Optional name for the span. If not provided, uses function name.
            variable_mapping: Optional mapping of expected parameter names to actual parameter names
            metadata_config: Optional list of metadata keys to include. If None, includes all metadata.

        Examples:
            # Case 1: Normal function with return value
            @general_tracing()
            async def my_function(param1, param2):
                result = some_operation(param1, param2)
                return result

            # Case 2: Function without return value using capture_result
            from observe_traces import capture_result

            @general_tracing()
            async def my_void_function(param1, param2):
                result = some_operation(param1, param2)
                capture_result(result)  # Capture result for tracing
                # No return statement

            # Case 3: Custom span name
            @general_tracing(name="Custom Operation Name")
            async def my_function():
                pass

            # Case 4: Metadata filtering
            @general_tracing(metadata_config=["functionName", "timeTaken"])
            async def my_function():
                pass
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                tracer = GeneralTracer(
                    span_name=name,
                    variable_mapping=variable_mapping,
                    metadata_config=metadata_config,
                )
                return await tracer.trace(func, *args, **kwargs)

            return wrapper

        return decorator


# Global service instance
observability_service = ObservabilityService()


# Convenience functions that maintain backward compatibility
def llm_tracing(
    provider: str,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
    is_sdk: bool = False,
):
    """Backward compatible LLM tracing decorator."""
    return observability_service.llm_tracing(
        provider, variable_mapping, metadata_config, is_sdk
    )


def llm_streaming_tracing(
    provider: str,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
    is_sdk: bool = False,
):
    """Backward compatible LLM streaming tracing decorator."""
    return observability_service.llm_streaming_tracing(
        provider, variable_mapping, metadata_config, is_sdk
    )


def embedding_tracing(
    provider: str,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
):
    """Backward compatible embedding tracing decorator."""
    return observability_service.embedding_tracing(
        provider, variable_mapping, metadata_config
    )


def vectordb_tracing(
    provider: str,
    operation_type: str,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
):
    """Backward compatible vector DB tracing decorator."""
    return observability_service.vectordb_tracing(
        provider, operation_type, variable_mapping, metadata_config
    )


def reranking_tracing(
    provider: str,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
):
    """Backward compatible reranking tracing decorator."""
    return observability_service.reranking_tracing(
        provider, variable_mapping, metadata_config
    )


def general_tracing(
    name: Optional[str] = None,
    variable_mapping: Optional[Dict[str, str]] = None,
    metadata_config: Optional[List[str]] = None,
):
    """
    General-purpose decorator for tracing any function.

    Args:
        name: Optional name for the span. If not provided, uses function name.
        variable_mapping: Optional mapping of expected parameter names to actual parameter names
        metadata_config: Optional list of metadata keys to include. If None, includes all metadata.

    Examples:
        # Case 1: Normal function with return value
        @general_tracing()
        async def my_function(param1, param2):
            result = some_operation(param1, param2)
            return result

        # Case 2: Function without return value using capture_result
        from observe_traces import capture_result

        @general_tracing()
        async def my_void_function(param1, param2):
            result = some_operation(param1, param2)
            capture_result(result)  # Capture result for tracing
            # No return statement

        # Case 3: Custom span name
        @general_tracing(name="Custom Operation Name")
        async def my_function():
            pass

        # Case 4: Metadata filtering
        @general_tracing(metadata_config=["functionName", "timeTaken"])
        async def my_function():
            pass
    """
    return observability_service.general_tracing(
        name, variable_mapping, metadata_config
    )


def register_provider(
    provider_name: str, token_parser: Callable, response_extractor: Callable
):
    """Backward compatible provider registration for LLM providers."""
    from ..base.provider import LLMProvider

    class CustomLLMProvider(LLMProvider):
        def __init__(self):
            super().__init__(provider_name, response_extractor)

        def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
            return token_parser(response_data)

        def calculate_cost(
            self, tokens_data: Dict[str, int], model_name: str
        ) -> Dict[str, float]:
            # Default implementation - can be enhanced
            return {"input": 0.0, "output": 0.0, "total": 0.0}

    observability_service.register_llm_provider(
        provider_name, CustomLLMProvider()
    )


def register_embedding_provider(
    provider_name: str,
    token_parser: Callable,
    price_calculator: Callable,
    embeddings_extractor: Callable,
):
    """Backward compatible provider registration for embedding providers."""
    from ..base.provider import EmbeddingProvider

    class CustomEmbeddingProvider(EmbeddingProvider):
        def __init__(self):
            super().__init__(provider_name, embeddings_extractor)

        def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
            return token_parser(response_data)

        def calculate_cost(
            self, tokens_data: Dict[str, int], model_name: str
        ) -> Dict[str, float]:
            return price_calculator(model_name, tokens_data)

    observability_service.register_embedding_provider(
        provider_name, CustomEmbeddingProvider()
    )


def register_reranking_provider(
    provider_name: str,
    token_parser: Callable,
    price_calculator: Callable,
    rerank_results_extractor: Callable,
):
    """Backward compatible provider registration for reranking providers."""
    from ..base.provider import RerankingProvider

    class CustomRerankingProvider(RerankingProvider):
        def __init__(self):
            super().__init__(provider_name, rerank_results_extractor)

        def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
            return token_parser(response_data)

        def calculate_cost(
            self, tokens_data: Dict[str, int], model_name: str
        ) -> Dict[str, float]:
            return price_calculator(model_name, tokens_data)

    observability_service.register_reranking_provider(
        provider_name, CustomRerankingProvider()
    )



"""
Concrete embedding provider implementations.
"""

from typing import Any, Dict, List

from ..base.provider import EmbeddingProvider


class PineconeEmbeddingProvider(EmbeddingProvider):
    """Pinecone embedding provider implementation."""

    def __init__(self):
        super().__init__(
            name="pinecone",
            embeddings_extractor=self._extract_pinecone_embeddings,
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Pinecone embedding response."""
        if "usage" in response_data:
            usage = response_data.get("usage", {})
            return {"tokens": usage.get("total_tokens", 0)}
        return {"tokens": 0}

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Pinecone embedding model."""
        pricing = {
            "llama-text-embed-v2": 0.16,
            "multilingual-e5-large": 0.08,
            "pinecone-sparse-english-v0": 0.08,
        }

        tokens = tokens_data.get("tokens", 0)
        model_price = pricing.get(model_name, 0.0)
        total_price = (tokens / 1000000) * model_price

        return {
            "tokens": tokens,
            "price_per_1M": model_price,
            "total": total_price,
        }

    def _extract_pinecone_embeddings(
        self, data: Dict[str, Any]
    ) -> List[List[float]]:
        """Extract embeddings from Pinecone response."""
        return [item["values"] for item in data.get("data", [])]


class CohereEmbeddingProvider(EmbeddingProvider):
    """Cohere embedding provider implementation."""

    def __init__(self):
        super().__init__(
            name="cohere", embeddings_extractor=self._extract_cohere_embeddings
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Cohere embedding response."""
        meta = response_data.get("meta", {})
        billed_units = meta.get("billed_units", {})
        return {"tokens": billed_units.get("input_tokens", 0)}

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Cohere embedding model."""
        pricing = {
            "embed-english-v3.0": 0.1,
            "embed-multilingual-v3.0": 0.1,
        }

        tokens = tokens_data.get("tokens", 0)
        model_price = pricing.get(model_name, 0.0)
        total_price = (tokens / 1000000) * model_price

        return {
            "tokens": tokens,
            "price_per_1M": model_price,
            "total": total_price,
        }

    def _extract_cohere_embeddings(
        self, data: Dict[str, Any]
    ) -> List[List[float]]:
        """Extract embeddings from Cohere response."""
        return data.get("embeddings", {}).get("float", [])


class JinaEmbeddingProvider(EmbeddingProvider):
    """Jina embedding provider implementation."""

    def __init__(self):
        super().__init__(
            name="jina", embeddings_extractor=self._extract_jina_embeddings
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from Jina embedding response."""
        usage = response_data.get("usage", {})
        return {"tokens": usage.get("total_tokens", 0)}

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for Jina embedding model."""
        pricing = {
            "jina-embeddings-v2-base-en": 0.05,
            "jina-embeddings-v3": 0.12,
            "jina-embeddings-v2-base-code": 0.05,
        }

        tokens = tokens_data.get("tokens", 0)
        model_price = pricing.get(model_name, 0.0)
        total_price = (tokens / 1000000) * model_price

        return {
            "tokens": tokens,
            "price_per_1M": model_price,
            "total": total_price,
        }

    def _extract_jina_embeddings(
        self, data: Dict[str, Any]
    ) -> List[List[float]]:
        """Extract embeddings from Jina response."""
        return [item["embedding"] for item in data.get("data", [])]


class VoyageAIEmbeddingProvider(EmbeddingProvider):
    """VoyageAI embedding provider implementation."""

    def __init__(self):
        super().__init__(
            name="voyageai",
            embeddings_extractor=self._extract_voyageai_embeddings,
        )

    def parse_tokens(self, response_data: Dict[str, Any]) -> Dict[str, int]:
        """Parse tokens from VoyageAI embedding response."""
        usage = response_data.get("usage", {})
        return {"tokens": usage.get("total_tokens", 0)}

    def calculate_cost(
        self, tokens_data: Dict[str, int], model_name: str
    ) -> Dict[str, float]:
        """Calculate cost for VoyageAI embedding model."""
        pricing = {
            "voyage-3": 0.06,
            "voyage-3-lite": 0.02,
            "voyage-finance-2": 0.12,
            "voyage-law-2": 0.12,
            "voyage-code-2": 0.12,
            "voyage-code-3": 0.18,
            "voyage-3-large": 0.18,
        }

        tokens = tokens_data.get("tokens", 0)
        model_price = pricing.get(model_name, 0.0)
        total_price = (tokens / 1000000) * model_price

        return {
            "tokens": tokens,
            "price_per_1M": model_price,
            "total": total_price,
        }

    def _extract_voyageai_embeddings(
        self, data: Dict[str, Any]
    ) -> List[List[float]]:
        """Extract embeddings from VoyageAI response."""
        return [item["embedding"] for item in data.get("data", [])]



from typing import Any, Dict, List, Optional, Tuple


def get_mapped_param_value(
    params: Dict[str, Any],
    expected_key: str,
    variable_mapping: Dict[str, str] = None,
):
    """
    Get parameter value using variable mapping if provided, otherwise use expected key.

    Args:
        params: The function parameters dictionary
        expected_key: The expected parameter key (e.g., "system_prompt", "user_prompt")
        variable_mapping: Optional mapping of expected keys to actual parameter names

    Returns:
        The parameter value or None if not found
    """
    if variable_mapping and expected_key in variable_mapping:
        actual_key = variable_mapping[expected_key]
        return params.get(actual_key)
    return params.get(expected_key)


def process_system_prompt(system_prompt: Any) -> Tuple[str, Dict[str, Any]]:
    """
    Process system prompt and extract cache control metadata if present.

    Args:
        system_prompt: The system prompt which can be in various formats:
                      - String format: passed as-is
                      - List format: [{"type": "text", "text": "prompt", "cache_control": {...}}]

    Returns:
        Tuple of (processed_text, cache_metadata)
    """
    cache_metadata = {}

    # Handle list format with cache_control
    if isinstance(system_prompt, list):
        processed_text = ""
        for item in system_prompt:
            if isinstance(item, dict):
                # Extract text content
                if item.get("type") == "text" and "text" in item:
                    processed_text += item["text"]

                # Extract cache_control metadata
                if "cache_control" in item:
                    cache_metadata["systemPromptCacheControl"] = item[
                        "cache_control"
                    ]

        return processed_text, cache_metadata

    # Handle other formats (string, etc.) - pass as-is
    return system_prompt or "", cache_metadata


def convert_anthropic_messages_to_langfuse_format(
    messages: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """
    Convert Anthropic messages to Langfuse format.
    """
    return [
        {"role": msg["role"], "content": msg["content"]} for msg in messages
    ]


def convert_groq_messages_to_langfuse_format(
    messages: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """
    Convert Groq messages to Langfuse format.
    """
    return [
        {"role": msg["role"], "content": msg["content"]} for msg in messages
    ]


def extract_tool_calls_from_messages(
    messages: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """
    Extract tool calls from the messages list.

    Args:
        messages: List of message dictionaries in Anthropic format

    Returns:
        List of tool call dictionaries
    """
    tool_calls = []

    for message in messages:
        if message.get("role") == "assistant":
            content = message.get("content", [])

            # Handle both string content and list content
            if isinstance(content, str):
                continue  # String content doesn't contain tool calls
            elif isinstance(content, list):
                for content_block in content:
                    if (
                        isinstance(content_block, dict)
                        and content_block.get("type") == "tool_use"
                    ):
                        tool_call = {
                            "id": content_block.get("id", ""),
                            "name": content_block.get("name", ""),
                            "input": content_block.get("input", {}),
                            "type": "tool_use",
                        }
                        tool_calls.append(tool_call)

    return tool_calls


def extract_tool_results_from_messages(
    messages: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """
    Extract tool results from the messages list.

    Args:
        messages: List of message dictionaries in Anthropic format

    Returns:
        List of tool result dictionaries
    """
    tool_results = []

    for message in messages:
        if message.get("role") == "user":
            content = message.get("content", [])

            # Handle both string content and list content
            if isinstance(content, str):
                continue  # String content doesn't contain tool results
            elif isinstance(content, list):
                for content_block in content:
                    if (
                        isinstance(content_block, dict)
                        and content_block.get("type") == "tool_result"
                    ):
                        tool_result = {
                            "tool_use_id": content_block.get("tool_use_id", ""),
                            "content": content_block.get("content", ""),
                            "is_error": content_block.get("is_error", False),
                            "type": "tool_result",
                        }
                        tool_results.append(tool_result)

    return tool_results


def get_tool_call_summary_from_messages(
    messages: List[Dict[str, Any]],
) -> Dict[str, Any]:
    """
    Get a summary of tool calls and tool results from the messages list.

    Args:
        messages: List of message dictionaries in Anthropic format

    Returns:
        Dictionary containing tool call and tool result information
    """
    tool_calls = extract_tool_calls_from_messages(messages)
    tool_results = extract_tool_results_from_messages(messages)

    return {
        "tool_calls": tool_calls,
        "tool_results": tool_results,
        "has_tool_calls": len(tool_calls) > 0,
        "has_tool_results": len(tool_results) > 0,
        "tool_call_count": len(tool_calls),
        "tool_result_count": len(tool_results),
    }


def extract_tool_names_from_tools(tools: List[Dict[str, Any]]) -> List[str]:
    """
    Extract tool names from the tools list.
    """
    return [tool.get("name", "Not Found") for tool in tools]


def filter_metadata(
    metadata: Dict[str, Any], metadata_config: Optional[List[str]] = None
) -> Dict[str, Any]:
    """
    Filter metadata dictionary to only include keys specified in metadata_config.

    Args:
        metadata: The complete metadata dictionary
        metadata_config: Optional list of keys to include. If None, returns all metadata.

    Returns:
        Filtered metadata dictionary
    """
    if metadata_config is None:
        return metadata

    # Return only the keys that are in metadata_config and exist in metadata
    return {
        key: value for key, value in metadata.items() if key in metadata_config
    }


def log_error(error: Exception, context: str = ""):
    """Standardized error logging for the package."""
    error_msg = f"[LLM-Observability Package] {context}: {type(error).__name__}: {error}"
    print(error_msg)
